'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
TO_TRACE_LOG = 'to_event_log'
TO_EVENT_LOG = 'to_event_log'
TO_EVENT_STREAM = 'to_event_stream'
TO_DATAFRAME = 'to_dataframe'
FROM_DATAFRAME = 'from_dataframe'

DF_TO_EVENT_LOG_1V = 'df_to_event_log_1v'
DF_TO_EVENT_LOG_NV = 'df_to_event_log_nv'

DEEPCOPY = "deepcopy"
STREAM_POSTPROCESSING = "stream_postprocessing"
